<?php

/*
  ----------------------------------------------------------------------
  AlternC - Web Hosting System
  Copyright (C) 2000-2013 by the AlternC Development Team.
  https://alternc.org/
  ----------------------------------------------------------------------
  LICENSE

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License (GPL)
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  To read the license please visit http://www.gnu.org/copyleft/gpl.html
  ----------------------------------------------------------------------
  Purpose of file: Manage LXC-based virtual machines for AlternC
  ----------------------------------------------------------------------
*/


/** 
 * Manage LXC-based virtual machines for AlternC
 */ 
class alternc_lxc {

  var $persistent = '/var/cache/alternc-lxc/persistent/'; // We put persistent per-server and per-user data there (ssh key mainly)
  var $cache = '/var/cache/alternc-lxc/'; // We put the VM there.
  var $baseimage  = '/var/lib/alternc-lxc/vm_disk.img'; // This is a file that can be read-only mounted and contains our base system
  var $hookdir = '/var/lib/alternc-lxc/hooks/'; // This folder contains shell scripts to launch at start/stop time.
  var $template  = '/etc/alternc-lxc/lxc-config.template'; // This is the LXC configuration file template we copy at start time.
  var $lockfile = '/var/run/alternc-lxc.lock';  // If this file exists, no VM starting is allowed.
  var $conffile = '/etc/alternc-lxc/config.json'; // This is the configuration file containing all VM and AlternC SERVERS allowed.

  var $conf = array();
  
  const LXC_ERR_STDERR=0;
  const LXC_ERR_SYSLOG=1;
  public $errmode=self::LXC_ERR_STDERR; // Where shall we spit our errors / logs ?
  private $aprio=array(LOG_CRIT => "Critical",
                       LOG_ERR => "Error",
                       LOG_WARNING  => "Warning",
                       LOG_INFO => "Info",
                       LOG_DEBUG => "Debug",
                       );


  /**
   * Constructor
   * check cache and basedir, and load the VM configuration 
   */
  function alternc_lxc() {
    if (! is_dir($this->cache)) {
      echo "Error: the cache directory does not exist\n";
      echo "Tip: create \"".$this->cache."\" directory\n";
      die(2);
    }
    if (! is_file($this->baseimage)) {
      echo "Error: the baseimage file does not exist\n";
      echo "Tip: install debootstrap and use /var/lib/alternc-lxc/alternc_lxc_initme.sh\n";
      die(2);
    }

    $this->loadconf();
  }


  /**
   * Find the first unused VM and return the conf id
   * if all = true, return the list of all the available VMs
   */
  function free_vm($all=false) {
    $rr= array();
    foreach ($this->conf['vm'] as $k => $v ) {
      if (! file_exists($this->vm_dir($k))) {
        if (! $all) return $k;
        $rr[] = $k;
      }
    }
    return $rr;
  }


  /**
   * List all the VM and their status
   */
  function listvm() {
    $this->log(LOG_INFO,"Listing VMs");
    echo "NAME\tSTATUS\tSERVER\tLOGIN\tUID\tSTARTED\n";
    foreach($this->conf['vm'] as $k=>$v) {
      $statusfile=$this->vm_dir($k)."/status";
      if (file_exists($statusfile)) {
	$status=@json_decode(file_get_contents($statusfile),true);
	echo "$k\tRUNNING\t".$status["server"]."\t".$status["login"]."\t".$status["uid"]."\t[".date("Y-m-d H:i:s",$status["starttime"])."]\n";
      } else {
	echo "$k\tAVAILABLE\t\t\t\t\n";
      }
    }
  }


  /**
   * disable vm creation
   */
  function lock() {
    return touch($this->lockfile);
  }


  /**
   * enable vm creation
   */
  function unlock () {
    @unlink($this->lockfile);
  }


  /**
   * tell if vm creation are locked
   */
  function is_locked() {
    return file_exists($this->lockfile);
  }


  /**
   * Return an array with the number of total and free vm
   */
  function monit() { 
    $this->log(LOG_INFO,"Monitoring VMs");
    $total = count($this->conf['vm']);
    $free = count($this->free_vm(true));

    $o = array('total' => $total, 'free' => $free);
    return $o;
  }


  /**
   * Return information of the VM that is running for $server and $login
   * may return an empty array, which means no VM is running for this Server & User
   */
  function status($server,$login) {
    $this->log(LOG_INFO,"Getting status for $server and $login");
    $obj = array('error'=>'-1', 'vm'=>'', 'hostname'=>'', 'msg'=>'');

    $server=strtolower($server);
    if ( 
	( $server != trim(strval($server))) ||
	( $key != trim(strval($key))) ||
	( $login != trim(strval($login))) ||
	( $pass  != trim(strval($pass)))  ||
	( $uid != intval($uid) ) ||
	( $uid < 2000 ) ) {
      $obj['error'] = 1;
      $obj['msg']="Bad parameters";
      return $obj;
    }
    
    return $o;
  }


  /** 
   * Get the name of the running VM of user $login for server $server
   * having secret key $key
   */ 
  function get_vm($server, $key, $login) {
    $obj = array('error'=>'-1', 'msg'=>'');

    $server=strtolower($server);
    if ( 
	( $server != trim(strval($server))) ||
	( $key != trim(strval($key))) ||
	( $login != trim(strval($login)))) {
      $obj['error'] = 1;
      $obj['msg']="Bad parameters";
      $this->log(LOG_ERR,"get_vm failed from server '$server' for login '$login', bad parameters");
      return $obj;
    }

    // Is the Server's key correct?
    if ($key!=$this->conf['servers'][$server]['key']) { 
      $obj['error'] = 6;
      $obj['msg']="Servers's key is incorrect, please ask your administrator.";
      $this->log(LOG_ERR,"get_vm failed from server '$server' for login '$login', incorrect server key");
      return $obj;
    }
    // Find the VM this user is running : 
    foreach($this->conf['vm'] as $k=>$v) {
      $vmdir=$this->vm_dir($k);
      if (is_file($vmdir."/status")) {
	$status=@json_decode(file_get_contents($vmdir."/status"),true);
	if ($status["server"]==$server && $status["login"]==$login) {
	  $obj = array(
		       'vm'=>$k, 
		       'hostname' => $v['hostname'], 
		       'starttime' => $status['starttime']
		       );
	  if (file_exists($this->cache."/persistent/".$server."/".substr($status['uid'],-1)."/".$status["uid"]."-".$status["login"]."/keys_fingerprint.txt")) {
	    $obj['ssh-keys']=explode("\n",file_get_contents($this->cache."/persistent/".$server."/".substr($status['uid'],-1)."/".$status["uid"]."-".$status["login"]."/keys_fingerprint.txt"));
	  }
	  $this->log(LOG_ERR,"get_vm running vm found for server '$server' and login '$login': $k");
	  return $obj;
	}
      }
    } 
    $obj['error'] = 7;
    $obj['msg']="No VM found for this user & server.";
    return $obj;
  } /* get_vm */


  /** 
   * Start a new virtual machine asked from alternc server $server
   * having secret key $key
   * for AlternC account $login having encrypted password $pass
   * and unix user-id $uid with $params parameters (for more customization)
   */ 
  function start_vm($server, $key, $login, $pass, $uid, $params = false) {
    $obj = array('error'=>'-1', 'vm'=>'', 'hostname'=>'', 'msg'=>'');

    $server=strtolower($server);
    if ( 
	( $server != trim(strval($server))) ||
	( $key != trim(strval($key))) ||
	( $login != trim(strval($login))) ||
	( $pass  != trim(strval($pass)))  ||
	( $uid != intval($uid) ) ||
	( $uid < 2000 ) ) {
      $obj['error'] = 1;
      $obj['msg']="Bad parameters";
      $this->log(LOG_ERR,"NOT STARTING VM from server '$server' for login '$login' and uid '$uid', bad parameters");
      return $obj;
    }
    $uid=intval($uid);

    if ($this->is_locked()) {
      $obj['error'] = 4;
      $obj['msg']="Virtual Machine are currently unavailable.";
      $this->log(LOG_ERR,"NOT STARTING VM from server '$server' for login '$login' and uid '$uid', virtual machines unavailable");
      return $obj;
    }

    // Find an available VM
    $vm = $this->free_vm();
    if (empty($vm)) {
      $obj['error'] = 2;
      $obj['msg']="no more Virtual Machine available, please try later.";
      $this->log(LOG_ERR,"NOT STARTING VM from server '$server' for login '$login' and uid '$uid', no more virtual machines available");
      return $obj;
    }

    // Is the Server's key correct?
    if ($key!=$this->conf['servers'][$server]['key']) { 
      $obj['error'] = 6;
      $obj['msg']="Servers's key is incorrect, please ask your administrator.";
      $this->log(LOG_ERR,"NOT STARTING VM from server '$server' for login '$login' and uid '$uid', incorrect server key");
      return $obj;
    }

    // ALL IS FINE, LET'S START A VM!
    $this->log(LOG_INFO,"STARTING VM from server '$server' for login '$login' and uid '$uid'");

    $vmdir = $this->vm_dir($vm);
    // Create working dir
    mkdir($vmdir);
    mkdir($vmdir."/rootfs");

    // We generate a status file in $vmdir/status
    $status=array(
		  "server" => $server,
		  "login" => $login,
		  "uid" => $uid,
		  "starttime"=>time(),
		  "started"=>date("Y-m-d H:i:s")
		  );
    if ($from=getenv("REMOTE_HOST")) { // from xinetd ;)
      $status["from"]=$from;
      $status["source"]="inetd";
    } else {
      if (list($from)=explode(" ",getenv("SSH_CLIENT"))) { // from ssh when using cli ;) 
	$status["from"]=$from;
	$status["source"]="cli";
      }
    }
    file_put_contents($vmdir."/status",json_encode($status,true) );

    if (! $this->generateconf($vm) ) {
      $obj['error'] = 3;
      $obj['msg']="no conf generated";
      $this->log(LOG_ERR,"FAILED while STARTING VM from server '$server' for login '$login' and uid '$uid', NO CONF generated");
      return $obj;
    }

    $exe="mount -o loop,ro ".$this->baseimage." ".escapeshellarg($vmdir."/rootfs")." 2>&1";
    exec($exe, $exec_output, $exec_return);
    $this->log(LOG_DEBUG,"mounting the rootfs, command was '$exe'");

    // If there are some error during mount
    if ($exec_return != 0 ) {
      $obj['error'] = 5;
      $obj['msg']="error during mount";
      if (isset($exec_output[0])) $obj['msg'].=" => ".$exec_output[0];
      $this->stop_vm($vm); // stop the vm
      $this->log(LOG_ERR,"FAILED while STARTING VM from server '$server' for login '$login' and uid '$uid', MOUNT failed");
      return $obj;
    }

    $this->launch_hooks("start", $vm, $server, $login, $pass, $uid);

    // Now we mount the home folder (if it exist in the config file)
    if (isset($this->conf['servers'][$server]['htmlmount']) && $this->conf['servers'][$server]['htmlmount']!="") {
      // no escapeshellarg on purpose: this allow a content like "-t nfs -o vers=3 10.1.2.3:/var/www/alternc" or sshfs ...
      $suffix="/".substr($login,0,1)."/".$login;
      $exe="mount ".$this->conf['servers'][$server]['htmlmount'].$suffix." ".escapeshellarg($vmdir."/rootfs/home/".$login."")." 2>&1";
      exec($exe,$exec_output,$exec_return);
      $this->log(LOG_DEBUG,"mounting the htmlhome, command was '$exe'");
      if ($exec_return!=0) {
	$obj['error'] = 5;
	$obj['msg']="error during mount HOME";
	if (isset($exec_output[0])) $obj['msg'].=" => ".$exec_output[0];
	exec("umount ".escapeshellarg($vmdir."/rootfs/home/".$login.""));
	exec("umount ".escapeshellarg($vmdir."/rootfs"));
	$this->stop_vm($vm); // stop the vm
	$this->log(LOG_ERR,"FAILED while STARTING VM from server '$server' for login '$login' and uid '$uid', MOUNT HOME failed");
	return $obj;
      }      
    }

    if (!is_dir("$vmdir/rootfs/home/$login/lxc-home")) {
      mkdir("$vmdir/rootfs/home/$login/lxc-home");
      file_put_contents("$vmdir/rootfs/home/$login/lxc-home/.htaccess","Deny from all\n");
    }
    if (!chown("$vmdir/rootfs/home/$login/lxc-home",$uid)) {
      $this->log(LOG_ERR,"FAILED doing chown of '$vmdir/rootfs/home/$login/lxc-home' into '$uid'");
    }
    if (!chgrp("$vmdir/rootfs/home/$login/lxc-home",$uid)) {
      $this->log(LOG_ERR,"FAILED doing chgrp of '$vmdir/rootfs/home/$login/lxc-home' into '$uid'");
    }
    if (!chown("$vmdir/rootfs/home/$login/lxc-home/.htaccess",$uid)) {
      $this->log(LOG_ERR,"FAILED doing chown of '$vmdir/rootfs/home/$login/lxc-home/.htaccess' into '$uid'");
    }
    if (!chgrp("$vmdir/rootfs/home/$login/lxc-home/.htaccess",$uid)) {
      $this->log(LOG_ERR,"FAILED doing chgrp of '$vmdir/rootfs/home/$login/lxc-home/.htaccess' into '$uid'");
    }

    // If $params, so we may have a few more informations.
    // Try to use it ;)

    // Set the default my.cnf for the user
    if ( !empty($params['mysql_host']) ) {
      // Special case: if we have a mysql_host set to 127.0.0.1 or ::1 or localhost, we use the LXC-SERVER'S config.json mysql_host and port instead !!!
      if ( $params['mysql_host']=="127.0.0.1" || $params['mysql_host']=="::1" || $params['mysql_host']=="localhost") {
	if (isset($this->conf['servers'][$server]['mysql_host'])) {
	  $params['mysql_host']=$this->conf['servers'][$server]['mysql_host'];
	  if (!isset($this->conf['servers'][$server]['mysql_port']))
	    $params['mysql_port']=3306;
	  else
	    $params['mysql_port']=intval($this->conf['servers'][$server]['mysql_port']);
	  file_put_contents("$vmdir/rootfs/home/$login/lxc-home/.my.cnf", "# WARNING: this file is overwritten each time you start a VM !\n[client]\nhost = ".$params['mysql_host']."\nport = ".$params['mysql_port']."\n");	
	}
	// else, use the server's provided MySQL_HOST and PORT.
      } else {
	if (!isset($params['mysql_port']))
	  $params['mysql_port']=3306;
	else
	  $params['mysql_port']=intval($params['mysql_port']);
	file_put_contents("$vmdir/rootfs/home/$login/lxc-home/.my.cnf", "# WARNING: this file is overwritten each time you start a VM !\n[client]\nhost = ".$params['mysql_host']."\nport = ".$params['mysql_port']."\n");
      }
    }
    if (file_exists("$vmdir/rootfs/home/$login/lxc-home/.my.cnf")) {
      chown("$vmdir/rootfs/home/$login/lxc-home/.my.cnf",$uid);
      chgrp("$vmdir/rootfs/home/$login/lxc-home/.my.cnf",$uid);
    }

    $exe="lxc-start -n $vm -f '$vmdir/config' -d -o '$vmdir/$vm.log'";
    exec($exe);
    $this->log(LOG_DEBUG,"starting the LXC, command was '$exe'");
    sleep(1); // let the log file be populated
    $exec2_output = file_get_contents("$vmdir/$vm.log");

    // If there are some error during launch
    if ( strpos($exec2_output, ' ERROR ')) {
      $obj['error'] = 5;
      $obj['msg']="error during launch => ".str_replace("\n", " -- ",$exec2_output);
      $this->stop_vm($vm); // stop the vm
      $this->log(LOG_ERR,"FAILED while STARTING VM from server '$server' for login '$login' and uid '$uid', error during launch: '".$exec2_output."'");
      return $obj;
    }

    $obj['error'] = 0;
    $obj['vm'] = $vm;
    $obj['hostname'] = $this->conf['vm'][$vm]['hostname'];
    $this->log(LOG_INFO,"SUCCESS, STARTING VM '$vm' from server '$server' for login '$login' and uid '$uid'");
    return $obj;
  }


  /** 
   * Stop all Virtual Machines
   * regardless of their AlternC server attachment
   */
  function stopall_vm() {
    $this->log(LOG_INFO,"STOP ALL VM called");
    $r = array();
    foreach ($this->conf['vm'] as $k => $v ) {
      $r[] = $this->stop_vm($k);
    }
    return $r;
  }


  /**
   * Stop the Virtual Machine $vm (set by its name)
   */
  function stop_vm($vm){
    $this->log(LOG_INFO,"STOPPING VM $vm");
    $obj = array('error'=>'-1', 'vm'=> '', 'msg'=>'' );

    if (! array_key_exists($vm, $this->conf['vm'])) {
      $obj['error'] = 0;
      $obj['msg'] = "This VM does not exist";
      return $obj;
    }

    exec("lxc-stop -n $vm ");
    sleep(1); // wait a little to let the VM stop

    $this->launch_hooks("stop", $vm);

    $vmdir=$this->vm_dir($vm);
    if ( is_dir( $vmdir."/rootfs")) { 
      if ( is_dir( $vmdir."/rootfs/etc") ) exec("umount $vmdir/rootfs");
      rmdir($vmdir."/rootfs");
    }
    if (file_exists($vmdir."/config")) unlink($vmdir."/config") ;
    if (file_exists($vmdir."/status")) unlink($vmdir."/status") ;
    if (file_exists($vmdir."/$vm.log")) unlink($vmdir."/$vm.log") ;
    if (is_dir($vmdir)) rmdir($vmdir);

    $obj = array('error'=>'0', 'vm'=> $vm, 'msg'=>'' );
    return $obj;

  }


  /* ------------------------------------------------------------ */
  /* private functions */
  /* ------------------------------------------------------------ */


  /** Compute AlternC's persistent home directory from a login */
  function persistentdir($server, $login, $uid) {
    return $this->persistent."/".$server."/".substr($uid,-1)."/".$uid."-".$login;
  }


  /** 
   * Log the $message with $priority (see priority list above)
   * to the current log destination (set with $this->errmode)
   */
  function log($priority,$message) {
    static $logopened=false;
    // we *always* log to syslog: 
    if (!$logopened) {
      $logopened=true;
      openlog("AlternC-LXC", LOG_NDELAY | LOG_PID, LOG_DAEMON);
    } 
    syslog($priority,$message);

    // we *also* log to stderr when requested:
    if ($this->errmode==self::LXC_ERR_STDERR) {
      $f=@fopen("php://stderr","ab");
      if ($f) {
        fputs($f,"[".date("Y-m-d H:i:s")."] ".$this->aprio[$priority].": ".str_replace("\n"," ",$message)."\n");
        fclose($f);
      }
    }
  }


  /**
   * replace die() function 
   * log the $message using our log function
   * and exit.
   */
  function mydie($message="") {
    if ($message=="") {
      $message="FATAL(1)";
      $code=1;
    }
    $code=1;
    $this->log(LOG_CRIT,$message);
    exit($code);
  }


  /**
   * Load the VM configuration file default in /etc/alternc-lxc/vm_config.json
   */
  function loadconf($cf=null){
    if (empty($cf)) $cf=$this->conffile;
    if (! file_exists($cf)) {
      $this->mydie("Error: missing configuration file $cf");
    }
    if (! $confjson=file_get_contents($cf) ) {
      $this->mydie("Error: can't load $cf");
    }
    if (! $this->conf=json_decode($confjson, true) ) {
      $this->mydie("Error: can't decode configuration in $cf");
    }
    // TODO: check that we have at least ONE server and ONE VM with at least ONE proper Network conf AND check that EACH network conf have a proper link+ip. mydie('error') else. 
  }


  /**
   * Replace in the data parameters all the "AlternC" magic var
   * (for example: %%VM%%, %%VMDIR%%, ...) 
   * Used to populate templates
   */
  function subvar($data, $vm) {
    $data = str_replace('%%VM%%',$vm,$data);
    $data = str_replace('%%VMDIR%%', $this->vm_dir($vm),$data);
    return $data;
  }


  /**
   * Return the VM directory
   */
  function vm_dir($vm) { 
    return rtrim($this->cache,"/")."/".$vm;
  }


  /**
   * Generate the LXC configuration file
   */
  function generateconf($vm) {
    $netip=array("ipv4","ipv4.gateway","ipv6","ipv6.gateway");
    $vmdir = $this->vm_dir($vm);
    file_put_contents($vmdir."/config", $this->subvar(file_get_contents($this->template),$vm));

    // Now generate the network conf
    foreach ( $this->conf['vm'][$vm]['network'] as $k => $v ) {

      if (!isset($v['mac'])) $v['mac']="42:42:42:".dechex(rand(0,15)).dechex(rand(0,15)).":".dechex(rand(0,15)).dechex(rand(0,15)).":".dechex(rand(0,15)).dechex(rand(0,15));

      $more=""; // This set IPv4 and/or IPv6 IP+GATEWAY network configuration 
      foreach($netip as $moreconf) {
	if (isset($v[$moreconf])) 
	  $more.="lxc.network.".$moreconf." = ".$v[$moreconf]."\n"; 
      }

      file_put_contents($vmdir."/config", 
			"# Auto generated for $k\n".
			"lxc.network.type = veth\n".
			"lxc.network.flags = up\n".
			"lxc.network.link = ".$v['link']."\n".
			"lxc.network.hwaddr = ".$v['mac']."\n".
			"lxc.network.name = $k\n".
			$more,
			FILE_APPEND);
    }
    return true;
  }


  /**
   * Launch all hooks in hook directory when VM are started or stopped
   */
  function launch_hooks($action, $vm, $server=null, $login=null, $pass=null, $uid=null) {
    if (! in_array($action,array("start", "stop"))) {
      $this->mydie("Error: action not allowed");
    }

    $ret=array();
    $c=opendir($this->hookdir);
    while ($di=readdir($c)) {
      $file = $this->hookdir."/".$di;
      if ( is_file($file) && is_executable($file) ){
        $ret[] = $file;
      }
    }
    closedir($c);
    asort($ret);
    foreach ( $ret as $rr ) {
      // called as: ./script start vmname vmdir login pass uid persistentdir
      $exe="$rr $action $vm ".escapeshellarg($this->vm_dir($vm))." ".escapeshellarg($server)." ".escapeshellarg($login).
	" ".escapeshellarg($pass)." $uid ".escapeshellarg($this->persistentdir($server,$login, $uid));
      exec($exe,$out,$ret);
      $this->log(LOG_INFO,"Launched hook '$exe', return code is '$ret', stdout is '".implode("|",$out)."'");
    }
  }
  

} // class alternc_lxc
?>
